"""
Module contains base settings classes
"""

from typing import Any, Dict, Optional, Union


class BaseSettingClass:
    """
    Base setting class
    Attributes:
        _serviceName(str): service name
        _groupName(str): name of group with current settings or None if there is only one setting in group
    """

    def __init__(self, serviceName: str, settingGroupName: Optional[str] = None):
        """
        Initialize
        Args:
            serviceName(str): service name
            settingGroupName(str): setting's group(if exists)
        """
        self._serviceName = serviceName
        self._groupName = settingGroupName

    @property
    def groupName(self) -> Union[str, None]:
        """Return group name"""
        return self._groupName

    def asDict(self) -> Dict[str, Any]:
        """
        Get class dict
        Returns:
            dict with all not-protected class members
        """
        return {key: value for key, value in self.__dict__.items() if not key.startswith("_")}

    def __repr__(self) -> str:
        """
        Class instance representation
        Returns:
            f"<{self.__class__.__name__}: {self._serviceName}.{self._groupName}>"
        """
        return f"<{self.__class__.__name__}: {self._serviceName}.{self._groupName}>"

    def __eq__(self, other) -> bool:
        """
        Operator equal

        Args:
            other: other settings

        Returns:
            True if dicts equals, otherwise False
        Raises:
            TypeError: if type 'other' is not BaseSettingClass
        """
        if isinstance(other, BaseSettingClass):
            return self.__dict__ == other.__dict__
        raise TypeError(f"Got wrong type {type(other)}.")
