"""
Module contains classes for migration
"""

from dataclasses import dataclass
from enum import Enum
from typing import Any, Callable, List, Optional

SETTING_TYPE = Any  # pylint: disable-msg=invalid-name


def getQuotedIfStr(value: Any):
    """Get value if not string else get value with quotes"""
    return f'"{value}"' if isinstance(value, str) else value


@dataclass
class LimitationToModify:  # pylint: disable-msg=too-many-instance-attributes
    """Limitation to modify data"""

    limitationName: str
    oldServices: List[str]
    newServices: List[str]
    oldValidationSchema: Optional[dict] = None
    newValidationSchema: Optional[dict] = None
    oldDefaultValue: Optional[Any] = None
    newDefaultValue: Optional[Any] = None
    oldDescription: Optional[str] = None
    newDescription: Optional[str] = None
    oldGroupNames: Optional[List[str]] = None
    newGroupNames: Optional[List[str]] = None
    settingUpgradeFunction: Optional[Callable[[SETTING_TYPE], SETTING_TYPE]] = None
    settingDowngradeFunction: Optional[Callable[[SETTING_TYPE], SETTING_TYPE]] = None

    @classmethod
    def fromDicts(cls, oldLimitation: dict, newLimitation: dict) -> "LimitationToModify":
        """Convert dict with old limitation and dict with new limitation to LimitationToModify"""
        return cls(
            limitationName=getQuotedIfStr(oldLimitation["limitation_name"]),
            oldServices=oldLimitation["services"],
            newServices=newLimitation["services"],
            oldValidationSchema=oldLimitation["validation_schema"],
            newValidationSchema=newLimitation["validation_schema"],
            oldDefaultValue=getQuotedIfStr(oldLimitation["default_value"]),
            newDefaultValue=getQuotedIfStr(newLimitation["default_value"]),
            oldDescription=getQuotedIfStr(oldLimitation["description"]),
            newDescription=getQuotedIfStr(newLimitation["description"]),
            oldGroupNames=oldLimitation["group_names"],
            newGroupNames=newLimitation["group_names"],
        )


@dataclass
class ExternalLimitationToModify:  # pylint: disable-msg=too-many-instance-attributes
    """ExternalLimitation to modify data"""

    limitationName: str
    oldServices: List[str]
    newServices: List[str]
    oldGroupNames: Optional[List[str]] = None
    newGroupNames: Optional[List[str]] = None

    @classmethod
    def fromDicts(cls, oldLimitation: dict, newLimitation: dict) -> "ExternalLimitationToModify":
        """Convert dict with old limitation and dict with new limitation to ExternalLimitationToModify"""
        return cls(
            limitationName=getQuotedIfStr(oldLimitation["limitation_name"]),
            oldServices=oldLimitation["services"],
            newServices=newLimitation["services"],
            oldGroupNames=oldLimitation["group_names"],
            newGroupNames=newLimitation["group_names"],
        )


@dataclass
class Group:
    """New group data"""

    groupName: str
    description: str

    @classmethod
    def fromDict(cls, data: dict) -> "Group":
        """Convert dict to group"""
        return cls(groupName=getQuotedIfStr(data["group_name"]), description=getQuotedIfStr(data["description"]))


@dataclass
class Limitation:
    """New limitation data"""

    limitationName: str
    services: List[str]
    validationSchema: dict
    defaultValue: Any
    description: str
    groupNames: List[str]

    @classmethod
    def fromDict(cls, data: dict) -> "Limitation":
        """Convert dict to limitation"""
        return cls(
            limitationName=getQuotedIfStr(data["limitation_name"]),
            services=data["services"],
            validationSchema=data["validation_schema"],
            defaultValue=getQuotedIfStr(data["default_value"]),
            description=getQuotedIfStr(data["description"]),
            groupNames=data["group_names"],
        )


class OperationMode(Enum):
    """Operation mode enum"""

    # pylint: disable-msg=invalid-name
    upgrade = 0
    downgrade = 1
