"""
Module contains migration generator
It requires valid old dump and available version path (see consts.py) to work properly
Also it depends on template to generate valid migration
It is possible to modify migration before launch, for more information see readme.md
"""

import decimal
import json
import secrets
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, KeysView, List

from black import FileMode, Report, TargetVersion, WriteBack, reformat_one
from mako.template import Template

from ..utils.dump import generateDump
from .classes import ExternalLimitationToModify, Group, Limitation, LimitationToModify
from .consts import (
    ACTUAL_DUMP_FILENAME,
    BLACK_LIST_SERVICES,
    BYTES_HASH_LENGTH,
    REVISION_PATH_NAME,
    TEMPLATE_NAME,
    VERSION_FIELD,
)
from .logger import Logger


class MigrationGenerator:
    """
    Migration generator
    - get old dump and its version from `actual_dump.json`
    - generate new dump from all_settings.py
    - compare the dumps and generate a migration
    - write down the migration to a `<revision name>.py` file, format using black
    - overwrite `actual_dump.json` with the new one
    """

    def __init__(self, userMessage: str, services: List[str], settingsStore: Dict[str, Any], migrationDir: Path):
        self.userMessage = userMessage.replace(" ", "_")
        self.logger = Logger(template="generator")
        self.possibleServices = services
        self.store = settingsStore
        self.migrationDir = migrationDir

    def getActualDumpFilePath(self) -> Path:
        """Get actual dump file name"""
        return self.migrationDir / ACTUAL_DUMP_FILENAME

    def getActualDump(self) -> dict:
        """Get actual dump file as dict"""
        return json.loads(self.getActualDumpFilePath().read_text(), parse_float=decimal.Decimal())  # type: ignore

    @staticmethod
    def getGroupsToAdd(oldGroupMap: dict, newGroups: KeysView) -> List[Group]:
        """Get groups to add"""
        return [Group.fromDict(data) for name, data in oldGroupMap.items() if name not in newGroups]

    @staticmethod
    def getLimitationsToAdd(oldLimitationMap: dict, newLimitations: KeysView) -> List[Limitation]:
        """Get limitations to add"""
        return [Limitation.fromDict(data) for name, data in oldLimitationMap.items() if name not in newLimitations]

    @staticmethod
    def getLimitationsToModify(oldLimitationMap: dict, newLimitationMap: dict) -> List[LimitationToModify]:
        """
        Get list of limitations to modify
        Args:
            oldLimitationMap: old map limitationName to limitation
            newLimitationMap: new map limitationName to limitation
        Returns
            all limitations which are different
        """
        res = []
        for limitationName in set(oldLimitationMap).intersection(set(newLimitationMap)):
            if oldLimitationMap[limitationName] == newLimitationMap[limitationName]:
                continue
            res.append(
                LimitationToModify.fromDicts(
                    oldLimitation=oldLimitationMap[limitationName], newLimitation=newLimitationMap[limitationName]
                )
            )
        return res

    @staticmethod
    def getExternalLimitationsToModify(
        oldLimitationMap: dict, newLimitationMap: dict
    ) -> List[ExternalLimitationToModify]:
        """
        Get list of limitations to modify
        Args:
            oldLimitationMap: old map limitationName to limitation
            newLimitationMap: new map limitationName to limitation
        Returns
            all limitations which are different or which are not present in one of the maps.
        """
        res = []
        for limitationName in set(oldLimitationMap).intersection(set(newLimitationMap)):
            if oldLimitationMap[limitationName] == newLimitationMap[limitationName]:
                continue
            res.append(
                ExternalLimitationToModify.fromDicts(
                    oldLimitation=oldLimitationMap[limitationName], newLimitation=newLimitationMap[limitationName]
                )
            )
        for limitationName in set(oldLimitationMap) - set(newLimitationMap):
            stub = dict(limitation_name=limitationName, group_names=[], services=[])
            externalLimitation = ExternalLimitationToModify.fromDicts(
                oldLimitation=oldLimitationMap[limitationName], newLimitation=stub
            )
            res.append(externalLimitation)
        for limitationName in set(newLimitationMap) - set(oldLimitationMap):
            stub = dict(limitation_name=limitationName, group_names=[], services=[])
            externalLimitation = ExternalLimitationToModify.fromDicts(
                oldLimitation=stub, newLimitation=newLimitationMap[limitationName]
            )
            res.append(externalLimitation)
        return res

    def replaceDump(self, newDump: dict, newRevision: str) -> None:
        """Write new actual dump file using all_settings.py"""
        newDump[VERSION_FIELD] = newRevision
        self.getActualDumpFilePath().write_text(f"{json.dumps(newDump, indent=4)}\n")
        self.logger.info("Actual dump has been replaced")

    @staticmethod
    def formatFileUsingBlack(filePath: Path) -> None:
        """Format file using black"""
        reformat_one(
            src=filePath,
            fast=True,
            write_back=WriteBack.YES,
            mode=FileMode({TargetVersion.PY37}),
            report=Report(check=True, quiet=True, verbose=False),
        )

    def generate(self) -> None:
        """Generate configs migration and refresh actual dump"""
        oldDump = self.getActualDump()
        # remove services which does not require migrations
        serviceNames = set(self.possibleServices) - BLACK_LIST_SERVICES
        newDump = generateDump(set(serviceNames), self.store)

        oldGroupMap = {row["group_name"]: row for row in oldDump["groups"]}
        newGroupMap = {row["group_name"]: row for row in newDump["groups"]}
        oldRegularLimitationMap = {row["limitation_name"]: row for row in oldDump["limitations"] if not row["external"]}
        newRegularLimitationMap = {row["limitation_name"]: row for row in newDump["limitations"] if not row["external"]}
        oldExternalLimitationMap = {row["limitation_name"]: row for row in oldDump["limitations"] if row["external"]}
        newExternalLimitationMap = {row["limitation_name"]: row for row in newDump["limitations"] if row["external"]}

        migrationDate = str(datetime.utcnow().date()).replace("-", "_")
        migrationHash = secrets.token_hex(BYTES_HASH_LENGTH)
        migrationFileName = f"{migrationDate}_{self.userMessage}_{migrationHash}.py"
        migrationFilePath = self.migrationDir / REVISION_PATH_NAME / migrationFileName

        templatePath = Path(__file__).parent / TEMPLATE_NAME
        migrationFilePath.write_text(
            Template(filename=templatePath.as_posix()).render(
                revision=migrationHash,
                downRevision=oldDump.get(VERSION_FIELD, ""),
                creationDatetime=datetime.utcnow(),
                groupsToAdd=self.getGroupsToAdd(newGroupMap, oldGroupMap.keys()),
                groupsToRemove=self.getGroupsToAdd(oldGroupMap, newGroupMap.keys()),
                limitationsToRemove=self.getLimitationsToAdd(oldRegularLimitationMap, newRegularLimitationMap.keys()),
                limitationsToAdd=self.getLimitationsToAdd(newRegularLimitationMap, oldRegularLimitationMap.keys()),
                limitationsToModify=self.getLimitationsToModify(oldRegularLimitationMap, newRegularLimitationMap),
                externalLimitationsToModify=self.getExternalLimitationsToModify(
                    oldExternalLimitationMap, newExternalLimitationMap
                ),
            )
        )
        self.formatFileUsingBlack(migrationFilePath)
        self.logger.info(f"{migrationFileName} has been generated")
        self.replaceDump(newDump, migrationHash)

    def init(self, generate: bool = False):
        """Init migration folder and first migration"""
        self.migrationDir.mkdir(exist_ok=True)
        (self.migrationDir / REVISION_PATH_NAME).mkdir()
        migrationHash = secrets.token_hex(BYTES_HASH_LENGTH)
        migrationDate = str(datetime.utcnow().date()).replace("-", "_")
        migrationFileName = f"{migrationDate}_init_{migrationHash}.py"
        migrationFilePath = self.migrationDir / REVISION_PATH_NAME / migrationFileName
        templatePath = Path(__file__).parent / TEMPLATE_NAME

        migrationFilePath.write_text(
            Template(filename=templatePath.as_posix()).render(
                revision=migrationHash,
                downRevision="",
                creationDatetime=datetime.utcnow(),
            )
        )
        self.formatFileUsingBlack(migrationFilePath)

        newDump = generateDump(set(), {})
        self.replaceDump(newDump, migrationHash)
        self.logger.info(f"Init configs migration has been generated, {migrationFileName}")
        if generate:
            self.generate()
