"""
Module contains SingleMigration and Migration classes
"""

import importlib
import sys
from importlib.util import spec_from_file_location
from operator import itemgetter
from pathlib import Path
from typing import Callable, Dict, List, Optional, Tuple

import ujson
from sqlalchemy import delete, func, insert, or_, select, update
from sqlalchemy.engine import Connection, create_engine
from sqlalchemy.exc import DatabaseError

from . import models
from .classes import OperationMode
from .consts import ACTUAL_DUMP_FILENAME, BYTES_HASH_LENGTH, FIRST_MIGRATION_REVISION
from .db_migrate_helper import DatabaseMigrationHelper, assertSchemas
from .logger import Logger


def importModule(modulePath: Path):
    """
    Import module from file
    Args:
        modulePath: module path.

    Returns:
        loaded module
    """
    spec = spec_from_file_location(modulePath.stem[-BYTES_HASH_LENGTH * 2 :], Path(modulePath))
    if spec is None:
        raise RuntimeError(f"failed to load a spec from {modulePath}")
    module = importlib.util.module_from_spec(spec)
    if spec.loader is None:
        raise RuntimeError(f"failed to load a spec from {modulePath}, loader is empty")
    spec.loader.exec_module(module)
    return module


class MigrationException(Exception):
    """Special exception occurred during migration"""


class SingleMigration:  # pylint: disable-msg=too-few-public-methods
    """Single migration - works with each userMessage"""

    def __init__(
        self,
        connection: Connection,
        actualRevision: str,
        operationMode: OperationMode,
        availableServices: List[str],
        versionsDir: Path,
        component: str,
    ):
        self.targetRevision: str
        self.migrateFunction: Callable

        self.logger = Logger(template=f"revision '{actualRevision}' {operationMode.name}")
        self.currentRevision: str = actualRevision
        self.operationMode: OperationMode = operationMode
        self.availableServices = availableServices
        self.dbMigrationHelper = DatabaseMigrationHelper(
            connection=connection,
            operationMode=self.operationMode,
            logger=self.logger,
            availableServices=self.availableServices,
        )
        self.versionsDir = versionsDir
        self.component = component
        self._initFromMigrationModule()

    def _initFromMigrationModule(self) -> None:
        """Init down revision, migration function and migration helper from current migration"""
        modulePath = next(self.versionsDir.glob(f"*{self.currentRevision}.py"))
        migrationModule = importModule(modulePath)
        migrationModule.mh = self.dbMigrationHelper
        if self.operationMode == OperationMode.upgrade:
            self.targetRevision = self.currentRevision
            self.migrateFunction = migrationModule.upgrade
        else:
            self.targetRevision = migrationModule.DOWN_REVISION
            self.migrateFunction = migrationModule.downgrade

    def migrate(self):
        """Migration database"""
        try:
            self.migrateFunction()
            self.dbMigrationHelper.connection.execute(
                update(models.ConfigsMigration)
                .values(revision=self.targetRevision)
                .where(models.ConfigsMigration.component == self.component)  # type: ignore
            )
        except DatabaseError as exc:
            raise RuntimeError(f"migration to '{self.targetRevision}' failed\n{exc}") from exc


class MigrationProcessor:
    """
    Migration processor
        - collect necessary revisions and determine migration order
        - run migration for each revision

    Attributes:
        engine: engine to configurator db
        services: interest services list. Migration applied for these services.
        migration: directory with configuration migration data
        component: component name. Union of settings for some services.
    """

    def __init__(
        self,
        dbUrl,
        migration: Path,
        component: str,
        services: List[str],
    ):
        self.engine = create_engine(dbUrl)  # pylint: disable-msg=C0103
        self.logger = Logger(template="migration")
        self.migration = migration
        self.services = services
        self.component = component
        sys.path.append(str(self.versionsDir))

    @property
    def versionsDir(self) -> Path:
        """Get folder with migration versions"""
        return self.migration / "versions"

    def getExistingRevisions(self) -> Dict[str, str]:
        """Get map with existing revisions and down revisions"""
        res = {}
        for fileName in self.versionsDir.glob("*.py"):
            migrationModule = importModule(fileName)
            if migrationModule.REVISION == FIRST_MIGRATION_REVISION:
                continue
            res[migrationModule.REVISION] = migrationModule.DOWN_REVISION
        return res

    @staticmethod
    def getSortedRevisions(oldToNewRevisionMap: Dict[str, str]) -> List[str]:
        """Get all sorted revisions"""
        if not oldToNewRevisionMap:
            return [FIRST_MIGRATION_REVISION]
        upRevisions = list(oldToNewRevisionMap.values())
        duplicatedHeads = {head for head, down in oldToNewRevisionMap.items() if upRevisions.count(down) > 1}
        if duplicatedHeads:
            raise MigrationException(f"Got duplicated heads revisions: {duplicatedHeads}")
        initialRevision = set(oldToNewRevisionMap.keys()).difference(oldToNewRevisionMap.values()).pop()
        currentRev: Optional[str] = initialRevision
        sortedRevisions = [initialRevision]
        while currentRev:
            nextRev = oldToNewRevisionMap.get(currentRev)
            if nextRev:
                sortedRevisions.append(nextRev)
            currentRev = nextRev
        return sortedRevisions[::-1]

    def getRevisionsToMigrate(
        self, oldToNewRevisionMap: Dict[str, str], currentRevision: str, targetRevision: Optional[str] = None
    ) -> Tuple[List[str], OperationMode, Optional[str]]:
        """Get migration revisions in prepared order and operation mode"""

        sortedRevisions = self.getSortedRevisions(oldToNewRevisionMap)
        currentRevIndex = sortedRevisions.index(currentRevision)

        if targetRevision == "head":
            targetRevIndex = len(sortedRevisions) - 1
            targetRevision = sortedRevisions[-1]
        elif targetRevision == "-1":
            if currentRevIndex == 0:
                targetRevIndex = None
                targetRevision = None
            else:
                targetRevIndex = max(0, currentRevIndex - 1)
                targetRevision = sortedRevisions[targetRevIndex]
        elif targetRevision == "null":
            targetRevIndex = None
            targetRevision = None
        elif targetRevision not in sortedRevisions:
            raise MigrationException(f"Target revision '{targetRevision}' not found")
        else:
            targetRevIndex = sortedRevisions.index(targetRevision)

        if targetRevIndex and currentRevIndex < targetRevIndex:
            revisionOrder = sortedRevisions[currentRevIndex + 1 : targetRevIndex + 1]
            operationMode = OperationMode.upgrade
        elif targetRevIndex is not None:
            revisionOrder = sortedRevisions[targetRevIndex + 1 : currentRevIndex + 1]
            revisionOrder.reverse()
            operationMode = OperationMode.downgrade
        else:
            revisionOrder = sortedRevisions[: currentRevIndex + 1]
            revisionOrder.reverse()
            operationMode = OperationMode.downgrade

        return revisionOrder, operationMode, targetRevision

    def getCurrentVersionFromDB(self) -> Optional[str]:
        """Get current revision from db"""
        with self.engine.begin() as connection:
            revision = connection.execute(
                select([models.ConfigsMigration.revision]).where(models.ConfigsMigration.component == self.component)
            ).scalar()

            if revision is None:
                self.logger.info(f"Revision not found for component {self.component}")
                return None
            self.logger.info(f"Current database configs revision is '{revision}'")
            return revision

    @property
    def dumpFile(self):
        """load dump file from disk"""
        return ujson.loads((self.migration / ACTUAL_DUMP_FILENAME).read_text())

    def _addExternalLimitationFromDump(self, limitation, dbContext):
        oldLimitation = dbContext.getLimitation(limitation["limitation_name"])
        limitationName = limitation["limitation_name"]
        currentSchema = oldLimitation["validation_schema"]
        expectedSchema = limitation["validation_schema"]

        assertSchemas(limitationName, currentSchema, expectedSchema, self.logger)

        dbContext.upgradeLimitationServices(limitationName, limitation["services"])
        # groups already created in fulfillDbFromTheDump
        dbContext.upgradeLimitationGroups(limitationName, limitation["group_names"])

    def fulfillDbFromTheDump(self, dump: dict, newComponent: bool = False) -> str:
        """
        Init default settings from the default settings dump `configs/migration/actual_dump.json`.

        Args:
            dump: loaded dump file
            newComponent: fill db for new component or not

        Returns:
            the dump-file revision
        """
        limitations, groups, revision = itemgetter("limitations", "groups", "version")(dump)

        with self.engine.begin() as connection:
            dbContext = DatabaseMigrationHelper(
                connection=connection,
                operationMode=OperationMode.upgrade,
                logger=self.logger,
                availableServices=self.services,
            )
            for group in groups:
                dbContext.addGroup(groupName=group["group_name"], groupDescription=group["description"])
            for limitation in limitations:
                if limitation["external"]:
                    self._addExternalLimitationFromDump(limitation, dbContext)
                else:
                    dbContext.addLimitation(
                        limitationName=limitation["limitation_name"],
                        services=limitation["services"],
                        validationSchema=limitation["validation_schema"],
                        defaultValue=limitation["default_value"],
                        description=limitation["description"],
                        groupNames=limitation["group_names"],
                    )
            if newComponent:
                connection.execute(
                    insert(models.ConfigsMigration).values(revision=revision, component=self.component)  # type: ignore
                )
            else:
                connection.execute(
                    update(models.ConfigsMigration)  # type: ignore
                    .values(revision=revision)
                    .where(models.ConfigsMigration.component == self.component)
                )

        return revision

    def applyDump(self, dump: dict, newComponent: bool = False) -> str:
        """Init configs migration table and migrate to last revision"""
        revision = self.fulfillDbFromTheDump(dump, newComponent=newComponent)
        self.logger.info(f"Empty database initialized with '{revision}' revision")
        return revision

    @staticmethod
    def lockTables(connection: Connection):
        """Lock all tables with data"""
        if connection.dialect.name == "postgresql":
            for model in [models.Limitation, models.Setting, models.Group, models.GroupLimitation, models.Tag]:
                connection.execute(f'LOCK TABLE "{model.__tablename__}" IN SHARE ROW EXCLUSIVE MODE;')  # type: ignore
        else:
            # oracle such oracle
            tableNames = [row[0] for row in connection.execute("select TABLE_NAME from USER_TABLES").cursor.fetchall()]
            for tableName in tableNames:
                connection.execute(f'LOCK TABLE "{tableName}" IN SHARE ROW EXCLUSIVE MODE')

    def isDbEmpty(self):
        """
        Check, whether there is no platform settings.

        Returns:
            database emptiness flag
        """
        selectSt = select([func.count(models.Limitation.limitation_name)]).where(
            or_(*(models.Limitation.services.like(f"%{serviceName}%") for serviceName in self.services))
        )
        with self.engine.begin() as connection:
            dbReply = connection.execute(selectSt).scalar()
            return not dbReply

    def initialMigration(self):
        """
        Create initial revision. Suppose that no

            1) add new component
            2) fill db from default dump
        """
        _, _, revision = itemgetter("limitations", "groups", "version")(self.dumpFile)
        self.applyDump(self.dumpFile, newComponent=True)
        return revision

    def _deleteComponent(self, connection: Connection):
        """
        Delete component from configs migration table
        Args:
            connection: connection to db
        """
        connection.execute(
            delete(models.ConfigsMigration).where(models.ConfigsMigration.component == self.component)  # type: ignore
        )

    def migrate(self, targetRevision: str) -> Optional[str]:
        """
        Run dump and/or db migration

        Args:
            targetRevision: target revision (head, -1, revision version, null)
        Returns:
            revision version after migration. if targetRevision is "null" (downgrade to initial state
            and remove component) will be return None.
        """
        currentRevision = self.getCurrentVersionFromDB()
        if currentRevision is None:
            currentRevision = self.initialMigration()

        elif currentRevision == FIRST_MIGRATION_REVISION:
            if self.isDbEmpty():
                currentRevision = self.applyDump(self.dumpFile)

        oldToNewRevisionMap = self.getExistingRevisions()
        revisions, operationMode, realTargetRevision = self.getRevisionsToMigrate(
            oldToNewRevisionMap=oldToNewRevisionMap,
            currentRevision=currentRevision,
            targetRevision=targetRevision,
        )
        if not revisions:
            self.logger.info(f"No migration required, already on '{targetRevision}'")
            return currentRevision
        with self.engine.begin() as connection:
            self.lockTables(connection)

            self.logger.info(f"Start migration: {operationMode.name} from {currentRevision}")
            for revision in revisions:
                SingleMigration(
                    connection=connection,
                    actualRevision=revision,
                    operationMode=operationMode,
                    availableServices=self.services,
                    versionsDir=self.versionsDir,
                    component=self.component,
                ).migrate()
            if targetRevision == "null":
                if realTargetRevision is not None:
                    raise RuntimeError(
                        f"User target revision is null, but estimated real target revision is not null: "
                        f"{realTargetRevision}"
                    )
                self._deleteComponent(connection)
                self.logger.info(f"Migration complete, component '{self.component}' was removed")
            else:
                if realTargetRevision is None:
                    self._deleteComponent(connection)
                    self.logger.info(f"Migration complete, component '{self.component}' was removed")
                else:
                    self.logger.info(f"Migration complete, current revision '{realTargetRevision}'")
            currentRevision = realTargetRevision

        return currentRevision
